"""This module handles `webnucleo <https://webnucleo.readthedocs.io>`_ collections of reactions."""

import math
import wnutils.xml as wx


class Reac:
    """A class for handling reactions and their data.

    Args:
        ``file`` (:obj:`str`): A string giving the XML file name with the\
           reaction data.

        ``reac_xpath`` (:obj:`str`, optional):  An XPath expression to\
         select reactions.  Default is all reactions.

    """

    def __init__(self, file, reac_xpath=""):
        self.xml = wx.Xml(file)
        self.reactions = {}
        self.reactions[""] = self.xml.get_reaction_data(reac_xpath=reac_xpath)

    def get_reactions(self, reac_xpath=""):
        """Method to return a collection of reactions.

        Args:
            ``reac_xpath`` (:obj:`str`, optional): An XPath expression\
             to select the reactions.  Default is all reactions in the\
             underlying collection of reactions.

        Returns:
            A :obj:`dict` containing `wnutils <https://wnutils.readthedocs.io>`_ reactions.

        """

        if reac_xpath not in self.reactions:
            self.reactions[reac_xpath] = {}

            new_reactions = self.xml.get_reaction_data(reac_xpath=reac_xpath)

            for key, value in new_reactions.items():
                if key in self.reactions[""]:
                    self.reactions[reac_xpath][key] = value

        return self.reactions[reac_xpath]

    def _compute_duplicate_factor(self, elements):
        my_dict = {}
        for species in elements:
            if species in my_dict:
                my_dict[species] += 1
            else:
                my_dict[species] = 1
        result = 1
        for val in my_dict.values():
            result *= math.factorial(val)
        return result

    def compute_reaction_duplicate_factors(self, name):
        """Method to compute the duplicate reaction element factors for a reaction.

        Args:
            ``name`` (:obj:`str`): A string giving the reaction.

        Returns:
            A two-element :obj:`tuple`.  The first element is the duplicate\
            factor for the forward reaction.  The second element is the\
            duplicate factor for the reverse reaction.

        """

        reaction = self.get_reactions()[name]
        return (
            self._compute_duplicate_factor(reaction.nuclide_reactants),
            self._compute_duplicate_factor(reaction.nuclide_products),
        )

    def compute_duplicate_factors(self, reac_xpath=""):
        """Method to compute the duplicate factors for reactions in the reaction collection.

        Args:
            ``reac_xpath`` (:obj:`str`, optional): An XPath expression\
            to select reactions.  Default is all reactions.

        Returns:
            A :obj:`dict` containing the factors.  The keys are reaction\
            strings.  The values are two-element :obj:`tuple` objects\
            with the first element the duplicate factor for the\
            forward reaction and the second element the duplicate factor\
            for the reverse reaction.

        """

        result = {}
        for reac in self.get_reactions(reac_xpath=reac_xpath):
            result[reac] = self.compute_reaction_duplicate_factors(reac)
        return result

    def is_weak_reaction(self, name):
        """Method to determine if a reaction is a weak reaction or not.

        Args:
            ``name`` (:obj:`str`): A string giving the reaction.

        Returns:
            A :obj:`bool` with value True if the reaction is a weak reaction\
            and False if not.

        """

        reaction = self.get_reactions()[name]
        result = False
        for _sp in reaction.reactants + reaction.products:
            if "electron" in _sp or "positron" in _sp or "neutrino" in _sp:
                result = True
        return result



"""This module handles `webnucleo <https://webnucleo.readthedocs.io>`_ collections of nuclides."""

import wnutils.xml as wx
import numpy as np
from scipy.interpolate import interp1d
import wnnet.consts as wc


class Nuc:
    """A class for handling nuclei and their data.

    Args:
        ``file`` (:obj:`str`): A string giving the XML file name with the\
           nuclide data.

        ``nuc_xpath`` (:obj:`str`, optional): An XPath expression to select\
        nuclides.  Default is all nuclides.

    """

    def __init__(self, file, nuc_xpath=""):
        self.xml = wx.Xml(file)
        self.nuclides = {}
        self.nuclides[""] = self.xml.get_nuclide_data(nuc_xpath=nuc_xpath)

    def get_nuclides(self, nuc_xpath=""):
        """Method to return a collection of nuclides.

        Args:
            ``nuc_xpath`` (:obj:`str`, optional): An XPath expression to select\
             the nuclides.  Default is all species in the underlying collection\
             of nuclides.

        Returns:
            A :obj:`dict` containing `wnutils <https://wnutils.readthedocs.io>`_ nuclides.

        """

        if nuc_xpath not in self.nuclides:
            self.nuclides[nuc_xpath] = {}

            new_nuclides = self.xml.get_nuclide_data(nuc_xpath=nuc_xpath)

            for key, value in new_nuclides.items():
                if key in self.nuclides[""]:
                    self.nuclides[nuc_xpath][key] = value

        return self.nuclides[nuc_xpath]

    def compute_nuclear_partition_function(self, name, t_9):
        """Method to compute the nuclear partition function for a species.

        Args:
            ``name`` (:obj:`str`): A string giving the name of the nuclide\
            whose partition function should be computed.

            ``t_9`` (:obj:`float`): The temperature in 10\\ :sup:`9` K at which\
            to compute the partition function.

        Returns:
            A :obj:`float` giving the nuclear partition function for the\
            species at the input temperature.

        """

        nuclide = self.get_nuclides()[name]

        v_t = nuclide["t9"]

        if len(v_t) == 0:
            return 2.0 * nuclide["spin"] + 1

        l_g = np.log10(nuclide["partf"])

        if t_9 < v_t[0]:
            return np.power(10.0, l_g[0])

        if t_9 > v_t[len(v_t) - 1]:
            return np.power(10.0, l_g[len(v_t) - 1])

        if len(v_t) <= 2:
            f_int = interp1d(v_t, l_g, kind="linear")
            return np.power(10.0, f_int(t_9))

        f_int = interp1d(v_t, l_g, kind="cubic")
        return np.power(10.0, f_int(t_9))

    def compute_quantum_abundance(self, name, t_9, rho):
        """Method to compute the quantum abundance of the nuclide at the
        input temperature and density.

        Args:
            ``name`` (:obj:`str`): The name of the species.

            ``t_9`` (:obj:`float`): The temperature in 10\\ :sup:`9` K at\
             which to compute the quantum abundance.

            ``rho`` (:obj:`float`): The density in g/cc  at which to compute\
            the quantum abundance.

        Returns:
            A :obj:`float` giving the quantum abundance for the species\
            at the input conditions.

        """

        assert t_9 > 0 and rho > 0, "t_9 and rho must be > 0."

        mass = self.compute_atomic_mass(name)

        result = self.compute_nuclear_partition_function(name, t_9) / (
            rho * wc.N_A
        )

        p_1 = (mass * wc.MeV_to_ergs) * wc.k_B * t_9 * 1.0e9
        p_2 = 2.0 * np.pi * np.power(wc.hbar * wc.c, 2)

        result *= np.power((p_1 / p_2), 1.5)

        return result

    def compute_atomic_mass(self, name):
        """Method to compute the atomic mass of a species.

        Args:
            ``name`` (:obj:`str`): The name of the species.

        Returns:
            A :obj:`float` containing the atomic mass in MeV.

        """
        nuclide = self.get_nuclides()[name]
        return nuclide["mass excess"] + nuclide["a"] * wc.m_u_in_MeV

    def compute_binding_energy(self, name):
        """Method to compute the nuclear binding energy of a species.

        Args:
            ``name`` (:obj:`str`): The name of the species.

        Returns:
            A :obj:`float` containing the binding energy in MeV.

        """

        nuclides = self.get_nuclides()
        nuclide = nuclides[name]
        delta_p = nuclides["h1"]["mass excess"]
        delta_n = nuclides["n"]["mass excess"]

        return (
            nuclide["z"] * delta_p
            + (nuclide["a"] - nuclide["z"]) * delta_n
            - nuclide["mass excess"]
        )

    def compute_nse_factor(self, name, t_9, rho):
        """Method to compute the NSE factor for a species.

        Args:
            ``name`` (:obj:`str`): The name of the species.

            ``t_9`` (:obj:`float`): The temperature in 10\\ :sup:`9` K at\
             which to compute the NSE factor.

            ``rho`` (:obj:`float`): The density in g/cc  at which to compute\
            the NSE factor.

        Returns:
            A :obj:`float` giving the NSE factor for the species\
            at the input conditions.  This factor is the binding energy\
            of the species divided by kT plus the logarithm of the\
            quantum abundance of the species.

        """

        return np.log(self.compute_quantum_abundance(name, t_9, rho)) + (
            (self.compute_binding_energy(name) * wc.MeV_to_ergs)
        ) / (wc.k_B * (t_9 * 1.0e9))



"""This module computes various reaction flows in a network."""

from dataclasses import dataclass
import functools
import numpy as np


@dataclass
class _FlowData:
    valid_reactions: dict
    dups: dict
    user_funcs: list
    scale: float = None
    direction: str = None
    order: str = None


def _set_user_funcs_for_zone(user_funcs, my_zone):
    result = {}
    if user_funcs:
        for key, value in user_funcs.items():
            result[key] = functools.partial(value, zone=my_zone)
    return result


def _compute_flows_for_valid_reactions(
    net, t_9, rho, mass_fractions, flow_data: _FlowData
):

    result = {}

    for reaction in flow_data.valid_reactions:
        _reaction = flow_data.valid_reactions[reaction]

        forward, reverse = net.compute_rates_for_reaction(
            reaction, t_9, user_funcs=flow_data.user_funcs
        )

        forward *= np.power(rho, len(_reaction.nuclide_reactants) - 1)
        forward /= flow_data.dups[reaction][0]
        forward *= _compute_abundance_product(
            net, mass_fractions, _reaction.nuclide_reactants
        )

        if not net.is_weak_reaction(reaction):
            reverse *= np.power(rho, len(_reaction.nuclide_products) - 1)
            reverse /= flow_data.dups[reaction][1]
            reverse *= _compute_abundance_product(
                net, mass_fractions, _reaction.nuclide_products
            )
        else:
            reverse = 0

        result[reaction] = (forward, reverse)

    return result


def compute_flows(net, t_9, rho, mass_fractions, **kwargs):
    """A routine to compute flows for a given set of mass fractions at the\
       input temperature and density.

    Args:
        ``net``: A wnnet network.

        ``t_9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
          to compute the flows.

        ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
          flows.

        ``mass_fractions`` (:obj:`float`):\
        A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
        fractions.

        ``**kwargs``: Allowed optional keyword arguments:

           *  **nuc_xpath** (:obj:`str`): XPath expression to select nuclides\
               for flow computations. Defaults to all species.
           *  **reac_xpath** (:obj:`str`): XPath expression to select reactions\
               for flow computations. Defaults to all reactions.
           *  **user_funcs** (:obj:`dict`): A dictionary of user-defined\
               functions associated with a user_rate key.\
               The prototype for each\
               user rate function should be (*reaction*, *t_9*), where\
               *t_9* is the temperature in billions of Kelvin and *reaction*\
               is a `wnutils <https://wnutils.readthedocs.io>`_ reaction\
               instance.  Other data can be bound to the function.

    Returns:
        A :obj:`dict` of reactions with each
        item in the dictionary a tuple giving the forward and
        reverse flow.

    """

    def get_args(**kwargs):
        my_defs = {
            "nuc_xpath": "",
            "reac_xpath": "",
            "user_funcs": "",
        }
        for key in kwargs:
            assert key in my_defs, f"{key} is not an allowed keyword"

        return {**my_defs, **kwargs}

    my_args = get_args(**kwargs)

    valid_reactions = net.get_valid_reactions(
        nuc_xpath=my_args["nuc_xpath"], reac_xpath=my_args["reac_xpath"]
    )

    dups = net.compute_duplicate_factors()

    flow_data = _FlowData(valid_reactions, dups, my_args["user_funcs"])

    return _compute_flows_for_valid_reactions(
        net,
        t_9,
        rho,
        mass_fractions,
        flow_data,
    )


def compute_flows_for_zones(
    net, zones, nuc_xpath="", reac_xpath="", user_funcs=""
):
    """A routine to compute flows for a set of zones.

    Args:
        ``net``: A wnnet network.

        ``zones`` (:obj:`dict`): A dictionary of\
         `wnutils <https://wnutils.readthedocs.io>`_ *zone data*.

        ``nuc_xpath`` (:obj:`str`, optional): XPath expression\
        to select nuclides for flow computations.  Defaults to all\
        species.

        ``reac_xpath`` (:obj:`str`, optional): XPath expression\
        to select reactions for flow computations.  Defaults to all\
        reactions.

        ``user_funcs`` (:obj:`dict`, optional): A dictionary of user-defined\
        functions associated with a user_rate key.\
        The prototype for each\
        user rate function should be (*reaction*, *t_9*, *zone*), where\
        *t_9* is the temperature in billions of Kelvin and *reaction* and\
        *zone* are `wnutils <https://wnutils.readthedocs.io>`_ reaction and
        zone instances.  Other data can be bound to the function.


    Returns:
        A :obj:`dict` of flows for each zone.  The data for
        each zone are themselves a :obj:`dict` of reactions with each
        item in the dictionary a tuple giving the forward and
        reverse flow.

    """

    zone_flows = {}

    valid_reactions = net.get_valid_reactions(
        nuc_xpath=nuc_xpath, reac_xpath=reac_xpath
    )

    dups = net.compute_duplicate_factors()

    for key, value in zones.items():
        props = value["properties"]
        if "t9" in props and "rho" in props:
            flow_data = _FlowData(
                valid_reactions,
                dups,
                _set_user_funcs_for_zone(user_funcs, value),
            )

            zone_flows[key] = _compute_flows_for_valid_reactions(
                net,
                float(props["t9"]),
                float(props["rho"]),
                value["mass fractions"],
                flow_data,
            )

    return zone_flows


def _swap_tuple_array(tup_array):
    new_array = []
    for tup in tup_array:
        new_array.append((tup[1], tup[0], tup[2]))
    return new_array


def _compute_link_flows_for_valid_reactions(
    net,
    t_9,
    rho,
    mass_fractions,
    flow_data: _FlowData,
):
    link_flows = {}

    for reaction in flow_data.valid_reactions.values():
        tup_array = []

        forward, reverse = net.compute_rates_for_reaction(
            reaction.get_string(), t_9, user_funcs=flow_data.user_funcs
        )

        if flow_data.direction in ("forward", "both"):
            forward *= np.power(rho, len(reaction.nuclide_reactants) - 1)
            forward /= flow_data.dups[reaction.get_string()][0]

            for i, source in enumerate(reaction.nuclide_reactants):
                p_source = _compute_abundance_product(
                    net,
                    mass_fractions,
                    reaction.nuclide_reactants,
                    exclude_index=i,
                )
                for target in reaction.nuclide_products:
                    tup_array.append(
                        (
                            source,
                            target,
                            forward * p_source * flow_data.scale,
                        )
                    )

                for target in reaction.nuclide_reactants:
                    tup_array.append(
                        (
                            source,
                            target,
                            -forward * p_source * flow_data.scale,
                        )
                    )

        if not net.is_weak_reaction(
            reaction.get_string()
        ) and flow_data.direction in (
            "reverse",
            "both",
        ):
            reverse *= np.power(rho, len(reaction.nuclide_products) - 1)
            reverse /= flow_data.dups[reaction.get_string()][1]

            for i, source in enumerate(reaction.nuclide_products):
                p_source = _compute_abundance_product(
                    net,
                    mass_fractions,
                    reaction.nuclide_products,
                    exclude_index=i,
                )
                for target in reaction.nuclide_reactants:
                    tup_array.append(
                        (
                            source,
                            target,
                            reverse * p_source * flow_data.scale,
                        )
                    )

                for target in reaction.nuclide_products:
                    tup_array.append(
                        (
                            source,
                            target,
                            -reverse * p_source * flow_data.scale,
                        )
                    )

        if flow_data.order == "reversed":
            tup_array = _swap_tuple_array(tup_array)

        link_flows[reaction.get_string()] = tup_array

    return link_flows


def compute_link_flows(net, t_9, rho, mass_fractions, **kwargs):
    """A routine to compute link flows for a given set of mass fractions\
       at the input temperature and density.

    Args:
        ``net``: A wnnet network.

        ``t_9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
          to compute the flows.

        ``rho`` (:obj:`float`):  The density in g/cc at which to compute\
          the flows.

        ``mass_fractions`` (:obj:`float`):\
          A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
          fractions.

        ``**kwargs``: Allowed optional keyword arguments:

           *  **nuc_xpath** (:obj:`str`): XPath expression to select nuclides\
               for flow computations. Defaults to all species.
           *  **reac_xpath** (:obj:`str`): XPath expression to select reactions\
               for flow computations. Defaults to all reactions.
           *  **user_funcs** (:obj:`dict`): A dictionary of user-defined\
              functions associated with a user_rate key.\
              The prototype for each\
              user rate function should be (*reaction*, *t_9*), where\
              *t_9* is the temperature in billions of Kelvin and *reaction*\
              is a `wnutils <https://wnutils.readthedocs.io>`_ reaction\
              instance.  Other data can be bound to the function.
           *  **direction** (:obj:`str`):  A string indicating the\
              direction of the links ("forward", from reactants to reactants\
              and products; "reverse", from products to reactants and products;\
              "both", both "forward" and "reverse").  Default is "both".
           *  **order** (:obj:`str`):  A string indicating the order\
              of the links.  Default is *normal*, in which the *source* and\
              *target* of the link are in the time-forward direction of the\
              reaction.  For *reversed*, the *source* and *target* are in the\
              opposite of the time-forward direction of the reaction such that]\
              the *target* is the *contribution* to the *source* over some\
              interval.

    Returns:
        A :obj:`dict` of reactions with each
        item in the dictionary an array of three-element :obj:`tuple` objects.
        The tuple elements are the *source*, *target*, and *link flow*.
        The *source* and *target* are determined from the time-forward
        direction of the reaction or the reverse, as indicated by the *order*
        argument.

    """

    def get_args(**kwargs):
        my_defs = {
            "nuc_xpath": "",
            "reac_xpath": "",
            "user_funcs": "",
            "direction": "both",
            "order": "normal",
        }
        for key in kwargs:
            assert key in my_defs, f"{key} is not an allowed keyword"

        return {**my_defs, **kwargs}

    my_args = get_args(**kwargs)

    assert my_args["direction"] in ("forward", "reverse", "both", "all")
    assert my_args["order"] in ("normal", "reversed")

    valid_reactions = net.get_valid_reactions(
        nuc_xpath=my_args["nuc_xpath"], reac_xpath=my_args["reac_xpath"]
    )

    dups = net.compute_duplicate_factors()

    scale = 1

    flow_data = _FlowData(
        valid_reactions,
        dups,
        my_args["user_funcs"],
        scale,
        my_args["direction"],
        my_args["order"],
    )

    return _compute_link_flows_for_valid_reactions(
        net,
        t_9,
        rho,
        mass_fractions,
        flow_data,
    )


def compute_link_flows_for_zones(net, zones, **kwargs):
    """A routine to compute link flows for a set of zones.

    Args:
        ``net``: A wnnet network.

        ``zones`` (:obj:`dict`): A dictionary of\
          `wnutils <https://wnutils.readthedocs.io>`_ *zone data*.

        ``**kwargs``: Allowed optional keyword arguments:

           *  **nuc_xpath** (:obj:`str`): XPath expression to select nuclides\
               for flow computations. Defaults to all species.
           *  **reac_xpath** (:obj:`str`): XPath expression to select reactions\
               for flow computations. Defaults to all reactions.
           *  **user_funcs** (:obj:`dict`): A dictionary of user-defined\
              functions associated with a user_rate key.\
              The prototype for each user rate function should be (*reaction*,\
              *t_9*, *zone*), where *t_9* is the temperature in billions of\
              Kelvin and *reaction* and *zone* are\
              `wnutils <https://wnutils.readthedocs.io>`_ reaction and\
              zone instances.  Other data can be bound to the function.
           *  **direction** (:obj:`str`):  A string indicating the\
              direction of the links ("forward", from reactants to reactants\
              and products; "reverse", from products to reactants and products;\
              "both", both "forward" and "reverse").  Default is "both".
           *  **order** (:obj:`str`):  A string indicating the order\
              of the links.  Default is *normal*, in which the *source* and\
              *target* of the link are in the time-forward direction of the\
              reaction.  For *reversed*, the *source* and *target* are in the\
              opposite of the time-forward direction of the reaction such that]\
              the *target* is the *contribution* to the *source* over some\
              interval.
           *  **include_dt** (:obj:`bool`):  Boolean determining whether\
              to include the *dt* (time interval) in the flow (True) or\
              not (False).  Default is False.

    Returns:
        A :obj:`dict` of flow links for each zone.  The data for
        each zone are themselves a :obj:`dict` of reactions with each
        item in the dictionary an array of three-element :obj:`tuple` objects.
        The *source* and *target* are determined from the time-forward
        direction of the reaction or the reverse, as indicated by the *order*
        argument.

    """

    def get_args(**kwargs):
        my_defs = {
            "nuc_xpath": "",
            "reac_xpath": "",
            "user_funcs": "",
            "direction": "both",
            "include_dt": False,
            "order": "normal",
        }
        for key in kwargs:
            assert key in my_defs, f"{key} is not an allowed keyword"

        return {**my_defs, **kwargs}

    my_args = get_args(**kwargs)

    assert my_args["direction"] in ("forward", "reverse", "both")
    assert my_args["order"] in ("normal", "reversed")

    valid_reactions = net.get_valid_reactions(
        nuc_xpath=my_args["nuc_xpath"], reac_xpath=my_args["reac_xpath"]
    )

    dups = net.compute_duplicate_factors()

    zone_link_flows = {}

    for key, value in zones.items():
        props = value["properties"]
        if "t9" in props and "rho" in props:
            if my_args["include_dt"]:
                scale = float(props["dt"])
            else:
                scale = 1

            flow_data = _FlowData(
                valid_reactions,
                dups,
                _set_user_funcs_for_zone(my_args["user_funcs"], value),
                scale,
                my_args["direction"],
                my_args["order"],
            )

            zone_link_flows[key] = _compute_link_flows_for_valid_reactions(
                net,
                float(props["t9"]),
                float(props["rho"]),
                value["mass fractions"],
                flow_data,
            )

    return zone_link_flows


def _compute_abundance_product(net, _x, sp_array, exclude_index=None):
    nuclides = net.get_nuclides()
    result = 1
    for i, _sp in enumerate(sp_array):
        if i is not exclude_index:
            tup = net.xml.get_z_a_state_from_nuclide_name(_sp)
            key = (_sp, tup[0], tup[1])
            if key in _x:
                _y = _x[key] / nuclides[_sp]["a"]
            else:
                _y = 0
            result *= _y
    return result



"""
A package of python routines to handle webnucleo networks.
"""

import wnnet.consts
import wnnet.nuc
import wnnet.reac
import wnnet.net
import wnnet.graph
import wnnet.flows
import wnnet.thermo
import wnnet.zones



"""This module computes thermodynamic quantities for data in a nuclear
network."""

import numpy as np
import wnstatmech as ws
import wnnet.consts as wc


class Nuclei:
    """A class for computing thermodynamic quantities for a collection of
    nuclei.

    Args:
        ``nuc`` (:obj:`wnnet.nuc.Nuc`): A wnnet nuclide collection.

        ``y_cutoff`` (:obj:`float`, optional): The smallest abundance
        per nucleon to be used in thermodynamic quantity calculations.

    """

    def __init__(self, nuc, y_cutoff=1.0e-25):
        self.nuc = nuc
        self.functions = {}

        assert y_cutoff >= 0
        self.y_cut = y_cutoff

        self.functions["number density"] = self.default_number_density
        self.functions["pressure"] = self.default_pressure
        self.functions["entropy density"] = self.default_entropy_density
        self.functions["energy density"] = self.default_energy_density
        self.functions["internal energy density"] = (
            self.default_internal_energy_density
        )

    def _compute_abundances(self, mass_fractions):
        y = {}
        for key, value in mass_fractions.items():
            y_c = value / key[2]
            if y_c > self.y_cut:
                y[key[0]] = y_c
        return y

    def default_number_density(self, t9, rho, mass_fractions):
        """The default routine for computing the number density of nuclei.

        Args:
            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the number density.

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            number density.

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

        Returns:
            A :obj:`float` giving the number density in cgs units.

        """
        assert t9 > 0
        y = self._compute_abundances(mass_fractions)

        result = 0
        for val in y.values():
            result += rho * wc.N_A * val

        return result

    def default_pressure(self, t9, rho, mass_fractions):
        """The default routine for computing the pressure of nuclei.

        Args:
            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the pressure.

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            pressure.

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

        Returns:
            A :obj:`float` giving the pressure in cgs units.

        """
        return (
            wc.k_B
            * 1.0e9
            * t9
            * self.default_number_density(t9, rho, mass_fractions)
        )

    def default_entropy_density(self, t9, rho, mass_fractions):
        """The default routine for computing the entropy density of nuclei.

        Args:
            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the entropy density.

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            entropy density.

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

        Returns:
            A :obj:`float` giving the entropy density in cgs units.

        """
        y = self._compute_abundances(mass_fractions)

        result = 0
        for key, value in y.items():
            result += (
                value
                * rho
                * wc.N_A
                * wc.k_B
                * (
                    5.0 / 2.0
                    - np.log(
                        value
                        / self.nuc.compute_quantum_abundance(key, t9, rho)
                    )
                )
            )

        return result

    def default_energy_density(self, t9, rho, mass_fractions):
        """The default routine for computing the energy density of nuclei,
        which includes the rest-mass energy.

        Args:
            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the energy density (including the rest mass).

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            energy density (including the rest mass).

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

        Returns:
            A :obj:`float` giving the energy density in cgs units.

        """
        y = self._compute_abundances(mass_fractions)

        result = 0

        kt = wc.k_B * 1.0e9 * t9

        for key, value in y.items():
            result += (
                rho
                * wc.N_A
                * value
                * (
                    self.nuc.compute_atomic_mass(key) * wc.MeV_to_ergs
                    + 1.5 * kt
                )
            )

        return result

    def default_internal_energy_density(self, t9, rho, mass_fractions):
        """The default routine for computing the energy density of nuclei,
        which does not include the rest-mass energy.

        Args:
            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the internal energy density.

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            internal energy density.

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

        Returns:
            A :obj:`float` giving the internal energy density in cgs units.

        """
        return (
            1.5
            * wc.k_B
            * 1.0e9
            * t9
            * self.default_number_density(t9, rho, mass_fractions)
        )

    def compute_quantity(self, quantity, t9, rho, mass_fractions):
        """Routine to compute the thermodynamic quantity for the nuclei.

        Args:
            ``quantity`` (:obj:`str`):  The thermodynamic quantity to compute.

            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the quantity.

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            quantity.

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

        Returns:
            A :obj:`float` giving the thermo quantity in cgs units.

        """
        assert quantity in self.functions

        return self.functions[quantity](t9, rho, mass_fractions)

    def compute_chemical_potentials(
        self, t9, rho, mass_fractions, include_rest_mass=False
    ):
        """Routine to compute the chemical potentials of the nuclei
        in units of kT.

        Args:
            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the chemical potentials.

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            chemical potentials.

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

            ``include_rest_mass`` (:obj:`bool`):\
                A boolean to choose whether to include the rest mass in the
                chemical potential.

        Returns:
            A :obj:`dict` with the name of each species as the key and the
            chemical potential (divide by kT) as the value.

        """
        y = self._compute_abundances(mass_fractions)

        result = {}
        for key, value in y.items():
            result[key] = np.log(
                value / self.nuc.compute_quantum_abundance(key, t9, rho)
            )
            if include_rest_mass:
                result[key] += (
                    self.nuc.compute_atomic_mass(key)
                    * wc.MeV_to_ergs
                    / (wc.k_B * 1.0e9 * t9)
                )

        return result

    def update_function(self, quantity, function):
        """Routine to update or add a thermodynamic function for the nuclei.

        Args:
            ``quantity`` (:obj:`str`):  The thermodynamic quantity.

            ``function``: A function with prototype *(t9, rho, mass_fractions)*
            to compute the thermodynamic quantity for the nuclei.

        Returns:
            On successful return, the function for the quantity has been
            updated or added.

        """
        self.functions[quantity] = function


class Thermo:
    """A class for computing thermodynamic quantities for data in a
    nuclear reaction network.

    Args:
        ``nuc`` (:obj:`wnnet.nuc.Nuc`): A wnnet nuclide collection.


    """

    def __init__(self, nuc):
        self.nuc = nuc
        self.fermions = {}
        self.bosons = {}
        self.fermions["electron"] = ws.fermion.create_electron()
        self.bosons["photon"] = ws.boson.create_photon()
        self.nuclei = Nuclei(nuc)
        self.number_density = {}

    def update_boson(self, name, boson):
        """Routine to update a boson.  If the boson does not already exist,
        it will be added.

        Args:
            ``name`` (:obj:`str`):  The name of the boson.

            ``boson``: The `wnstatmech <https://wnstatmech.readthedocs.io>`__
            boson to update.

        Returns:
            On successful return, the boson has been added or updated.

        """
        self.bosons[name] = boson

    def get_boson(self, name):
        """Routine to retrieve a boson.

        Args:
            ``name`` (:obj:`str`):  The name of the boson.

        Returns:
            The `wnstatmech <https://wnstatmech.readthedocs.io>`__ boson.

        """
        assert name in self.bosons
        return self.bosons[name]

    def remove_boson(self, name):
        """Routine to remove a boson.

        Args:
            ``name`` (:obj:`str`):  The name of the boson.

        Returns:
            On successful return, the boson has been removed.

        """
        assert name in self.bosons
        self.bosons.pop(name)

    def update_fermion(self, name, fermion):
        """Routine to update a fermion.  If the fermion does not already
        exists, it will be added.

        Args:
            ``name`` (:obj:`str`):  The name of the fermion.

            ``fermion``: The `wnstatmech <https://wnstatmech.readthedocs.io/>`__
            fermion to update.

        Returns:
            On successful return, the fermion has been added or updated.

        """
        self.fermions[name] = fermion

    def get_fermion(self, name):
        """Routine to retrieve a fermion.

        Args:
            ``name`` (:obj:`str`):  The name of the fermion.

        Returns:
            The `wnstatmech <https://wnstatmech.readthedocs.io/>`__ fermion.

        """
        assert name in self.fermions
        return self.fermions[name]

    def remove_fermion(self, name):
        """Routine to remove a fermion.

        Args:
            ``name`` (:obj:`str`):  The name of the fermion.

        Returns:
            On successful return, the fermion has been removed.

        """
        assert name in self.fermions
        self.fermions.pop(name)

    def set_number_density(self, particle, number_density):
        """Routine to update the number density for a particle.  Use this
        routine to set the number density of particles other than photons,
        electrons, and nuclei.

        Args:
            ``particle`` (:obj:`str`):  The name of the particle.

            ``number_density`` (:obj:`float`): The number density in
            cgs units.

        Returns:
            On successful return, the number density has been updated.

        """
        self.number_density[particle] = number_density

    def get_number_density(self, particle):
        """Routine to retrieve the number density for a particle.
        Use this routine for particles other than the default
        particles (photons, electrons, and nuclei).  For the default
        particles, compute the number density directly
        from the *compute_quantity* method.

        Args:
            ``particle`` (:obj:`str`):  The name of the particle.

        Returns:
            The number density of the particle in cgs units.

        """
        return self.number_density[particle]

    def compute_quantity(self, quantity, t9, rho, mass_fractions):
        """A routine to compute a thermodynamic quantity for a given set of\
        mass fractions at the input temperature and density.

        Args:
            ``quantity`` (:obj:`str`):  The thermodynamic quantity to compute.

            ``t9`` (:obj:`float`):  The temperature in 10\\ :sup:`9` K at which\
            to compute the quantity.

            ``rho`` (:obj:`float`):  The density in g/cc at which to compute the\
            quantity.

            ``mass_fractions`` (:obj:`float`):\
                A `wnutils <https://wnutils.readthedocs.io>`_ dictionary of mass\
                fractions.

        Returns:
            A :obj:`dict` with keys *electron*, *photon*, *baryon*,
            and *total* and the values being the computed quantity for
            each key.

        """

        result = {}

        temperature = 1.0e9 * t9

        # Baryons

        result["baryon"] = self.nuclei.compute_quantity(
            quantity, t9, rho, mass_fractions
        )

        # Fermions

        for name, fermion in self.fermions.items():
            if name == "electron":
                _n = 0
                for key, value in mass_fractions.items():
                    _n += rho * wc.N_A * key[1] * value / key[2]
            else:
                _n = self.get_number_density(fermion)

            alpha = fermion.compute_chemical_potential(temperature, _n)

            result[name] = fermion.compute_quantity(
                quantity, temperature, alpha
            )

        # Bosons

        for name, boson in self.bosons.items():
            if name == "photon":
                alpha = 0
            else:
                alpha = boson.compute_chemical_potential(
                    temperature, self.get_number_density(boson)
                )

            result[name] = boson.compute_quantity(quantity, temperature, alpha)

        # Total

        total = 0
        for value in result.values():
            total += value

        result["total"] = total

        return result

    def compute_quantity_in_zone(self, quantity, zone):
        """A routine to compute a thermodynamic quantity for the input zone.

        Args:
           ``quantity`` (:obj:`str`):  The thermodynamic quantity to compute.

           ``zone`` (:obj:`dict`): A `wnutils <https://wnutils.readthedocs.io>`_ *zone*.

        Returns:
            A :obj:`dict` keys *electron*, *photon*, *baryon*,
            and *total* and the values being the computed quantity for
            the zone.

        """

        return self.compute_quantity(
            quantity,
            float(zone["properties"]["t9"]),
            float(zone["properties"]["rho"]),
            zone["mass fractions"],
        )

    def compute_quantity_for_zones(self, quantity, zones):
        """A routine to compute a thermodynamic quantity for the input zones.

        Args:
            ``quantity`` (:obj:`str`):  The thermodynamic quantity to compute.

            ``zones`` (:obj:`dict`): A dictionary of
            `wnutils <https://wnutils.readthedocs.io>`_ *zone data*.

        Returns:
            A :obj:`dict` with the key being the zone label and the value
            being another dictionary with keys *electron*, *photon*, *baryon*,
            and *total* and the values being the computed quantity for
            each zone.

        """

        result = {}

        for key, value in zones.items():
            result[key] = self.compute_quantity_in_zone(
                quantity,
                value,
            )

        return result



"""This module handles zone data from `webnucleo <https://webnucleo.readthedocs.io>`_ files."""

import wnutils.xml as wx


class Zones:
    """A class for handling webnucleo zones.

    Args:
        ``file`` (:obj:`str`):  A string giving the XML file name with the\
         zone data.

    """

    def __init__(self, file):
        self.xml = wx.Xml(file)

    def get_zones(self, zone_xpath=""):
        """Method to return zones.

        Args:
            ``zone_xpath`` (:obj:`str`, optional):  An XPath expression to\
             select zones.  Default is all zones.

        Returns:
            A :obj:`dict` of `wnutils <https://wnutils.readthedocs.io>`_\
             zone data objects.

        """

        return self.xml.get_zone_data(zone_xpath=zone_xpath)

    def write(self, file, pretty_print=True, zone_xpath=""):
        """Method to write the zones to an xml file.

        Args:

           ``file`` (:obj:`str`): A string giving the name of output\
            xml file.

           ``pretty_print`` (:obj:`bool`, optional): If set to True,\
           routine outputs the xml in nice indented format.

           ``zone_xpath`` (:obj:`str`, optional):  An XPath expression to\
             select zones.  Default is all zones.

        Returns:
            On successful return, the underlying xml has been written\
            to ``file``.

        """

        zones = self.get_zones(zone_xpath=zone_xpath)

        new_xml = wx.New_Xml("zone_data")

        new_xml.set_zone_data(zones)
        new_xml.write(file, pretty_print=pretty_print)
